#include "dkq.h"
#include "global.h"
#include "globmat.h"
#include "genfile.h"
#include "node.h"
#include "element.h"
#include "intpoints.h"
#include <stdlib.h>
#include <math.h>



dkq::dkq (void)
{
  long i,j;

  //  number nodes on element
  nne=4;
  //  number of DOFs on element
  ndofe=12;
  //  number of strain/stress components
  tncomp=3;
  //  number of components for graphic purposes
  gncomp=4;
  //  number of functions approximated
  napfun=3;
  //  order of numerical integration of mass matrix
  intordmm=2;
  //  number of edges on element
  ned=4;
  //  number of nodes on one edge
  nned=2;
  //  order of numerical integration on element edges (boundaries)
  intordb=2;

  //  number of blocks (parts of geometric matrix)
  nb=1;

  //  number of strain/stress components
  ncomp = new long [nb];
  ncomp[0]=3;
  
  //  cumulative number of components approximated
  cncomp = new long [nb];
  cncomp[0]=0;
  
  //  number of integration points
  //  order of numerical integration of stiffness matrix
  nip = new long* [nb];
  intordsm = new long* [nb];
  for (i=0;i<nb;i++){
    nip[i] = new long [nb];
    intordsm[i] = new long [nb];
  }
  
  intordsm[0][0]=2;
  
  nip[0][0]=4;
  
  //  total number of integration points
  tnip=0;
  for (i=0;i<nb;i++){
    for (j=0;j<nb;j++){
      tnip+=nip[i][j];
    }
  }
  
  //  strain/stress state
  ssst = platek;

}

dkq::~dkq (void)
{
  long i;
  
  for (i=0;i<nb;i++){
    delete [] nip[i];
    delete [] intordsm[i];
  }
  delete [] nip;
  delete [] intordsm;
  
  delete [] cncomp;
  delete [] ncomp;
}



/**
   function prepares auxiliary data for quadrilateral plate element
   
   @param x,y - vectors of node coordinates
   @param l - %vector of lengths
   @param sx,sy - vectors of direction vectors components
   @param nx,ny - vectors of normal vectors components

   JK, 20. 10. 2019
*/
void dkq::auxdata (vector &x,vector &y,vector &l,vector &sx,vector &sy,vector &nx,vector &ny)
{
  
  sx[0]=x[1]-x[0];  sy[0]=y[1]-y[0];
  l[0]=sqrt(sx[0]*sx[0]+sy[0]*sy[0]);
  sx[0]/=l[0];      sy[0]/=l[0];
  nx[0]=sy[0];      ny[0]=0.0-sx[0];

  sx[1]=x[2]-x[1];  sy[1]=y[2]-y[1];
  l[1]=sqrt(sx[1]*sx[1]+sy[1]*sy[1]);
  sx[1]/=l[1];      sy[1]/=l[1];
  nx[1]=sy[1];      ny[1]=0.0-sx[1];

  sx[2]=x[3]-x[2];  sy[2]=y[3]-y[2];
  l[2]=sqrt(sx[2]*sx[2]+sy[2]*sy[2]);
  sx[2]/=l[2];      sy[2]/=l[2];
  nx[2]=sy[2];      ny[2]=0.0-sx[2];

  sx[3]=x[0]-x[3];  sy[3]=y[0]-y[3];
  l[3]=sqrt(sx[3]*sx[3]+sy[3]*sy[3]);
  sx[3]/=l[3];      sy[3]/=l[3];
  nx[3]=sy[3];      ny[3]=0.0-sx[3];
}

/**
   function approximates function defined by nodal values

   @param xi,eta - coordinates on element
   @param nodval - nodal values
   
   JK, 20. 10. 2019
*/
double dkq::approx (double xi,double eta,vector &nodval)
{
  double f;
  vector bf(ASTCKVEC(nne));
  
  bf_lin_4_2d (bf.a,xi,eta);
  
  scprd (bf,nodval,f);

  return f;
}


/**
   function assembles strain-displacement (geometric) %matrix
   
   @param gm - geometric %matrix
   @param xi,eta - natural coordinates
   @param x,y - coordinates of nodes on the element
   @param sx,sy - coordinates of direction vectors
   @param nx,ny - coordinates of normal vectors
   @param l - %vector of edge lengths
   @param jac - Jacobian
   
   JK, 20. 10. 2019
*/
void dkq::geom_matrix (matrix &gm,double xi, double eta,vector &x,vector &y,vector &sx,vector &sy,vector &nx,vector &ny,vector &l,double &jac)
{
  long i;
  vector dx(ASTCKVEC(ndofe*2)),dy(ASTCKVEC(ndofe*2));
  
  dx_bf_quad_dkq (dx.a,xi,eta,l.a,sx.a,sy.a,nx.a,ny.a);
  dy_bf_quad_dkq (dy.a,xi,eta,l.a,sx.a,sy.a,nx.a,ny.a);
  
  derivatives_2d (dx,dy,jac,x,y,xi,eta);

  for (i=0;i<ndofe;i++){
    gm[0][i] = dx[i];
    gm[1][i] = dy[ndofe+i];
    gm[2][i] = dx[ndofe+i] + dy[i];
  }

}


/**
   function computes stiffness %matrix of plate quadrilateral
   finite element based on discrete Kirchhoff theory
   
   @param eid - number of element
   @param ri,ci - row and column indices
   @param sm - stiffness %matrix
   @param x,y - vectors of nodal coordinates

   JK, 20. 10. 2019
*/
void dkq::stiffness_matrix (long eid,long ri,long ci,matrix &sm,vector &x,vector &y)
{
  long i,j,ipp;
  double xi,eta,ww1,ww2,jac,thick;
  ivector nodes(ASTCKIVEC(nne));
  vector l(ASTCKVEC(nne)),sx(ASTCKVEC(nne)),sy(ASTCKVEC(nne)),nx(ASTCKVEC(nne)),ny(ASTCKVEC(nne)),w,gp,t(ASTCKVEC(nne));
  matrix gm(ASTCKMAT(tncomp,ndofe)),d(ASTCKMAT(tncomp,tncomp));
  
  //  nodes on element
  Mt->give_elemnodes (eid,nodes);
  //  thickness in nodes
  Mc->give_thickness (eid,nodes,t);
  //  auxiliary vectors (direction vectors, normal vectors and edge lengths
  auxdata (x,y,l,sx,sy,nx,ny);
  
  fillm (0.0,sm);
  
  //  coordinates of integration points and their weights
  reallocv (intordsm[0][0],w);
  reallocv (intordsm[0][0],gp);
  gauss_points (gp.a,w.a,intordsm[0][0]);
  
  //  integration point id
  ipp=Mt->elements[eid].ipp[0][0];
  
  
  for (i=0;i<intordsm[0][0];i++){
    xi=gp[i];  ww1=w[i];
    for (j=0;j<intordsm[0][0];j++){
      eta=gp[j];  ww2=w[j];
      
      //  geometric matrix
      geom_matrix (gm,xi,eta,x,y,sx,sy,nx,ny,l,jac);
      
      //  matrix of stiffness of the material
      Mm->matstiff (d,ipp);
      
      //  thickness in integration point
      thick = approx (xi,eta,t);
      
      jac*=thick*thick*thick*ww1*ww2;
      
      //  contribution to the stiffness matrix of the element
      bdbjac (sm,gm,d,gm,jac);
      
      ipp++;
    }
  }

}

/**
   function computes stiffness %matrix of quadrilateral plate
   element based on the discrete Kirchhoff theory
   
   @param lcid - load case id
   @param eid - element id
   @param sm - stiffness %matrix
   
   JK, 20. 10. 2019
*/
void dkq::res_stiffness_matrix (long eid,matrix &sm)
{
  long transf;
  ivector nodes(ASTCKIVEC(nne));
  vector x(ASTCKVEC(nne)),y(ASTCKVEC(nne));
  
  //  node coordinates
  Mt->give_node_coord2d (x,y,eid);
  
  //  stiffness matrix
  stiffness_matrix (eid,0,0,sm,x,y);
  
}
